package generated.geldkarteAsymm.javacard.geldkarte;

import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Util;
/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	private byte decodeByte(byte[] in) {
		if (in[encoding_length++] != Code.BYTE) {

			SimpleComm.raiseError();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) {
		if (in[encoding_length++] != Code.BOOLEAN) {

			SimpleComm.raiseError();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public short encode(Codeable o, byte[] dest) {
		encodeInit(dest);

		try {
			encode(o);
		}

		catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		return encoding_length;

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	private void encodeShort(short o) {
		encodeDestination[encoding_length++] = Code.SHORT;
		Util.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeShort(short o) {
		encoding_length += 3;
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	private short decodeShort(byte[] in) {
		if (in[encoding_length++] != Code.SHORT)
			SimpleComm.raiseError();
		short res = Util.getShort(in, encoding_length);
		encoding_length += 2;
		return res;
	}

	public void encodeByteArray(byte[] o) {

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		Util.setShort(encodeDestination, encoding_length, (short) (o.length));
		encoding_length += 2;

		Util.arrayCopy(o, (short) 0, encodeDestination, encoding_length,
				(short) o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	private void decodeByteArrayInto(byte[] in, byte[] into) {
		if (in[encoding_length++] != Code.BYTEARRAY)
			SimpleComm.raiseError();
		short len = Util.getShort(in, encoding_length);
		encoding_length += 2;
		Util.arrayCopy(in, encoding_length, into, (short) 0, len);
		encoding_length += len;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) {

		switch (c.getCode()) {
			case Code.GETKEY :
				encodeGetKey((GetKey) c);
				break;
			case Code.SENDKEY :
				encodeSendKey((SendKey) c);
				break;
			case Code.PAYSIG :
				encodePaySig((PaySig) c);
				break;
			case Code.PAY :
				encodePay((Pay) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.PAYDATA :
				encodePayData((PayData) c);
				break;
			case Code.ENCDATAASYMM :
				encodeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();

		}
	}

	public Message decodeMessage(byte[] in, short offset, short expectedLength) {
		encoding_length = offset;
		Message m = null;
		try {
			switch (in[encoding_length]) {
				case Code.GETKEY :
					m = Store.newGetKey();
					decodeGetKeyInto(in, (GetKey) m);
					break;
				case Code.SENDKEY :
					m = Store.newSendKey();
					decodeSendKeyInto(in, (SendKey) m);
					break;
				case Code.PAYSIG :
					m = Store.newPaySig();
					decodePaySigInto(in, (PaySig) m);
					break;
				case Code.PAY :
					m = Store.newPay();
					decodePayInto(in, (Pay) m);
					break;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		if (m == null || encoding_length != (short) (expectedLength + offset))
			SimpleComm.raiseError();
		return m;
	}

	public void compSize(Codeable c) {
		switch (c.getCode()) {
			case Code.GETKEY :

				compSizeGetKey((GetKey) c);
				break;
			case Code.SENDKEY :

				compSizeSendKey((SendKey) c);
				break;
			case Code.PAYSIG :

				compSizePaySig((PaySig) c);
				break;
			case Code.PAY :

				compSizePay((Pay) c);
				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);
				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);
				break;
			case Code.PAYDATA :

				compSizePayData((PayData) c);
				break;
			case Code.ENCDATAASYMM :

				compSizeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);
				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);
				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();
				break;

		}
	}

	public PlainData decodePlainData(byte[] in, byte expectedType) {
		encoding_length = 0;
		PlainData p = null;
		try {
			if (in[encoding_length] != expectedType) {

				SimpleComm.raiseError();
			}
			switch (in[encoding_length]) {

				case Code.PAYDATA :
					p = Store.newPayData();
					decodePayDataInto(in, (PayData) p);
					break;

			}
		} catch (ArrayIndexOutOfBoundsException e) {

			SimpleComm.raiseError();
		}

		if (p == null) {

			SimpleComm.raiseError();
		}
		return p;
	}

	private void encodeGetKey(GetKey c) {
		encodeDestination[encoding_length++] = Code.GETKEY;
	}

	private void decodeGetKeyInto(byte[] in, GetKey into) {
		if (in[encoding_length++] != Code.GETKEY)
			SimpleComm.raiseError();
	}

	private void compSizeGetKey(GetKey c) {
		encoding_length++;
	}

	private void encodeSendKey(SendKey c) {
		encodeDestination[encoding_length++] = Code.SENDKEY;
		encode(c.cert);
		encode(c.sig);
	}

	private void decodeSendKeyInto(byte[] in, SendKey into) {
		if (in[encoding_length++] != Code.SENDKEY)
			SimpleComm.raiseError();
		decodeCertificateInto(in, (Certificate) (into.cert));
		decodeSignedDataInto(in, (SignedData) (into.sig));
	}

	private void compSizeSendKey(SendKey c) {
		encoding_length++;
		compSize(c.cert);
		compSize(c.sig);
	}

	private void encodePaySig(PaySig c) {
		encodeDestination[encoding_length++] = Code.PAYSIG;
		encode(c.sig);
	}

	private void decodePaySigInto(byte[] in, PaySig into) {
		if (in[encoding_length++] != Code.PAYSIG)
			SimpleComm.raiseError();
		decodeSignedDataInto(in, (SignedData) (into.sig));
	}

	private void compSizePaySig(PaySig c) {
		encoding_length++;
		compSize(c.sig);
	}

	private void encodePay(Pay c) {
		encodeDestination[encoding_length++] = Code.PAY;
		encode(c.data);
	}

	private void decodePayInto(byte[] in, Pay into) {
		if (in[encoding_length++] != Code.PAY)
			SimpleComm.raiseError();
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.data));
	}

	private void compSizePay(Pay c) {
		encoding_length++;
		compSize(c.data);
	}

	private void encodeCertificate(Certificate c) {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;
		encodeByteArray((byte[]) c.name);
		encode(c.key);
	}

	private void decodeCertificateInto(byte[] in, Certificate into) {
		if (in[encoding_length++] != Code.CERTIFICATE)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.name);
		decodePublicKeyInto(in, (PublicKey) (into.key));
	}

	private void compSizeCertificate(Certificate c) {
		encoding_length++;
		compSizeByteArray(c.name);
		compSize(c.key);
	}

	private void encodeSignedData(SignedData c) {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;
		encodeByteArray((byte[]) c.signed);
	}

	private void decodeSignedDataInto(byte[] in, SignedData into) {
		if (in[encoding_length++] != Code.SIGNEDDATA)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.signed);
	}

	private void compSizeSignedData(SignedData c) {
		encoding_length++;
		compSizeByteArray(c.signed);
	}

	private void encodePayData(PayData c) {
		encodeDestination[encoding_length++] = Code.PAYDATA;
		encodeShort((short) c.amount);
		encode(c.challenge);
	}

	private void decodePayDataInto(byte[] in, PayData into) {
		if (in[encoding_length++] != Code.PAYDATA)
			SimpleComm.raiseError();
		into.amount = decodeShort(in);
		decodeNonceInto(in, (Nonce) (into.challenge));
	}

	private void compSizePayData(PayData c) {
		encoding_length++;
		compSizeShort(c.amount);
		compSize(c.challenge);
	}

	private void encodeEncDataAsymm(EncDataAsymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATAASYMM;
		encodeByteArray(c.encrypted);

		encodeShort(c.plainlength);
		encodeShort(c.enclength);

		encodeByte(c.plainDataType);
	}

	private void decodeEncDataAsymmInto(byte[] in, EncDataAsymm into) {
		if (in[encoding_length++] != Code.ENCDATAASYMM)
			SimpleComm.raiseError();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		decodeByteArrayInto(in, into.encrypted);
		into.plainlength = decodeShort(in);
		into.enclength = decodeShort(in);
		byte b = decodeByte(in);
		into.plainDataType = b;
		if (b != Code.PAYDATA)
			SimpleComm.raiseError();
	}

	private void compSizeEncDataAsymm(EncDataAsymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	private void encodePublicKey(PublicKey c) {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;

		encodeByteArray(c.modulus);
	}

	private void decodePublicKeyInto(byte[] in, PublicKey into) {
		if (in[encoding_length++] != Code.PUBLICKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		into.updateKey();
	}

	private void compSizePublicKey(PublicKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);

	}

	private void encodeNonce(Nonce c) {
		encodeDestination[encoding_length++] = Code.NONCE;
		encodeByteArray((byte[]) c.nonce);
	}

	private void decodeNonceInto(byte[] in, Nonce into) {
		if (in[encoding_length++] != Code.NONCE)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.nonce);
	}

	private void compSizeNonce(Nonce c) {
		encoding_length++;
		compSizeByteArray(c.nonce);
	}

	public void encodePrivateKey(PrivateKey c) {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.modulus);
		encodeByteArray(c.privexponent);

	}

	private void decodePrivateKeyInto(byte[] in, PrivateKey into) {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		decodeByteArrayInto(in, into.privexponent);
		into.updateKey();
	}

	private void compSizePrivateKey(PrivateKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);
		compSizeByteArray(c.privexponent);

	}

	public void decodeGeldkarte(byte[] in, short offset, Geldkarte thisGeldkarte) {
		encoding_length = offset;
		if (in[encoding_length++] != Code.GELDKARTE)
			SimpleComm.raiseError();

		if (in[encoding_length] != Code.IGNORE) {
			thisGeldkarte.balance = decodeShort(in);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeSignedDataInto(in, thisGeldkarte.cert);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePrivateKeyInto(in, thisGeldkarte.privKey);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeCertificateInto(in, thisGeldkarte.pubKey);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			thisGeldkarte.state = decodeByte(in);
		} else
			encoding_length++;
	}

}
