package generated.geldkarteAsymm.terminal.merchant;

import swt.util.ByteArray;
import java.nio.charset.Charset;

/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	private static Charset charset = Charset.forName("UTF-8");

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	/**
	 * decodes an object beginning at given offset
	 * 
	 * @param in array containing encoded object
	 * @param offset where the encoded object begins
	 * @return the decoded object
	 */

	private byte decodeByte(byte[] in) throws TerminalException {
		if (in[encoding_length++] != Code.BYTE) {

			stop();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) throws TerminalException {
		if (in[encoding_length++] != Code.BOOLEAN) {
			stop();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public Codeable decode(byte[] in, short offset) throws Exception {
		encoding_length = offset;
		return decode(in);
	}
	public byte[] stringToByte(String s) {
		byte[] b = s.getBytes(charset);
		byte[] ret = new byte[LengthConstants.LENGTHOFSTRING];
		int len = (b.length < LengthConstants.LENGTHOFSTRING)
				? b.length
				: LengthConstants.LENGTHOFSTRING;
		ByteArray.copy(b, 0, ret, 0, len);
		return ret;
	}

	public boolean isValidCardString(String s) {
		return (s.getBytes(charset).length <= LengthConstants.LENGTHOFSTRING);
	}

	public void encodeString(String i) {
		byte[] o = stringToByte(i);
		encodeDestination[encoding_length++] = Code.BYTEARRAY;
		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;
		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);
		encoding_length += o.length;
	}

	private void compSizeString(String i) {
		byte[] o = stringToByte(i);
		encoding_length += 3;
		encoding_length += o.length;
	}

	public short encode(Codeable o, byte[] dest) throws Exception {
		encodeInit(dest);

		encode(o);
		return computeSize(o);

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	public void encodeInt(int i) {
		short o = (short) i;
		if (i != (int) o)
			System.err.println("int too big for cast into short.");
		encodeDestination[encoding_length++] = Code.SHORT;
		ByteArray.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	private void compSizeInt(int i) {
		short o = (short) i;
		encoding_length += 3;
	}
	public int decodeInt(byte[] in) {
		short res = ByteArray.getShort(in, ++encoding_length);
		encoding_length += 2;
		return (int) res;
	}

	public byte[] decodeByteArray(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		return dest;
	}

	public String byteToString(byte[] b) {
		int len = b.length;
		while (len > 0 && b[len - 1] == 0)
			len--;
		return new String(ByteArray.subarray(b, 0, len), charset);
	}

	public String decodeString(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		String str = byteToString(dest);
		return str;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	public byte[] encodeSingleString(String s) {
		byte[] o = stringToByte(s);
		byte[] dest = new byte[3 + o.length];
		int i = 0;
		dest[i++] = Code.BYTEARRAY;
		ByteArray.setShort(dest, i, (short) (o.length));
		i += 2;
		ByteArray.copy(o, 0, dest, i, o.length);
		i += o.length;
		return dest;
	}

	public void encodeByteArray(byte[] o) {

		//byte[] o = stringToByte(i);

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;

		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) throws Exception {

		switch (c.getCode()) {
			case Code.GETKEY :
				encodeGetKey((GetKey) c);
				break;
			case Code.SENDKEY :
				encodeSendKey((SendKey) c);
				break;
			case Code.PAYSIG :
				encodePaySig((PaySig) c);
				break;
			case Code.PAY :
				encodePay((Pay) c);
				break;
			case Code.U :
				encodeU((U) c);
				break;
			case Code.UPAY :
				encodeUPay((UPay) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.PAYDATA :
				encodePayData((PayData) c);
				break;
			case Code.ENCDATAASYMM :
				encodeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				stop();

		}
	}

	public Codeable decode(byte[] in) throws Exception {
		switch (in[encoding_length]) {
			case Code.GETKEY :
				return decodeGetKey(in);
			case Code.SENDKEY :
				return decodeSendKey(in);
			case Code.PAYSIG :
				return decodePaySig(in);
			case Code.PAY :
				return decodePay(in);
			case Code.U :
				return decodeU(in);
			case Code.UPAY :
				return decodeUPay(in);
			case Code.CERTIFICATE :
				return decodeCertificate(in);
			case Code.SIGNEDDATA :
				return decodeSignedData(in);
			case Code.PAYDATA :
				return decodePayData(in);
			case Code.ENCDATAASYMM :
				return decodeEncDataAsymm(in);
			case Code.PUBLICKEY :
				return decodePublicKey(in);
			case Code.NONCE :
				return decodeNonce(in);
			case Code.PRIVATEKEY :
				return decodePrivateKey(in);
				// never reached
			default :
				stop();
		}
		return null;
	}

	public void compSize(Codeable c) throws Exception {
		switch (c.getCode()) {
			case Code.GETKEY :

				compSizeGetKey((GetKey) c);

				break;
			case Code.SENDKEY :

				compSizeSendKey((SendKey) c);

				break;
			case Code.PAYSIG :

				compSizePaySig((PaySig) c);

				break;
			case Code.PAY :

				compSizePay((Pay) c);

				break;
			case Code.U :

				compSizeU((U) c);

				break;
			case Code.UPAY :

				compSizeUPay((UPay) c);

				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);

				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);

				break;
			case Code.PAYDATA :

				compSizePayData((PayData) c);

				break;
			case Code.ENCDATAASYMM :

				compSizeEncDataAsymm((EncDataAsymm) c);

				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);

				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);

				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);

				break;
			// never reached
			default :

				stop();

		}
	}

	public PlainData decodePlainData(byte[] in, byte expectedType)
			throws Exception {
		encoding_length = 0;
		PlainData p = null;
		try {
			if (in[encoding_length] != expectedType) {
				stop();

			}
			switch (in[encoding_length]) {

				case Code.PAYDATA :
					p = decodePayData(in);
					break;

			}
		} catch (ArrayIndexOutOfBoundsException e) {
			stop();

		}

		if (p == null) {
			stop();

		}
		return p;
	}

	public void encodeGetKey(GetKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.GETKEY;
	}

	public GetKey decodeGetKey(byte[] in) throws Exception {
		GetKey c = new GetKey();
		encoding_length++;
		return c;
	}

	private void compSizeGetKey(GetKey c) throws Exception {
		encoding_length++;
	}

	public void encodeSendKey(SendKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.SENDKEY;

		encode(c.getCert());

		encode(c.getSig());
	}

	public SendKey decodeSendKey(byte[] in) throws Exception {
		SendKey c = new SendKey();
		encoding_length++;
		c.setCert(((Certificate) decode(in)));
		c.setSig(((SignedData) decode(in)));
		return c;
	}

	private void compSizeSendKey(SendKey c) throws Exception {
		encoding_length++;
		compSize(c.getCert());
		compSize(c.getSig());
	}

	public void encodePaySig(PaySig c) throws Exception {
		encodeDestination[encoding_length++] = Code.PAYSIG;

		encode(c.getSig());
	}

	public PaySig decodePaySig(byte[] in) throws Exception {
		PaySig c = new PaySig();
		encoding_length++;
		c.setSig(((SignedData) decode(in)));
		return c;
	}

	private void compSizePaySig(PaySig c) throws Exception {
		encoding_length++;
		compSize(c.getSig());
	}

	public void encodePay(Pay c) throws Exception {
		encodeDestination[encoding_length++] = Code.PAY;

		encode(c.getData());
	}

	public Pay decodePay(byte[] in) throws Exception {
		Pay c = new Pay();
		encoding_length++;
		c.setData(((EncDataAsymm) decode(in)));
		return c;
	}

	private void compSizePay(Pay c) throws Exception {
		encoding_length++;
		compSize(c.getData());
	}

	public void encodeU(U c) throws Exception {
		encodeDestination[encoding_length++] = Code.U;
	}

	public U decodeU(byte[] in) throws Exception {
		U c = new U();
		encoding_length++;
		return c;
	}

	private void compSizeU(U c) throws Exception {
		encoding_length++;
	}

	public void encodeUPay(UPay c) throws Exception {
		encodeDestination[encoding_length++] = Code.UPAY;

		encodeInt((int) c.getValue());
	}

	public UPay decodeUPay(byte[] in) throws Exception {
		UPay c = new UPay();
		encoding_length++;
		c.setValue(decodeInt(in));
		return c;
	}

	private void compSizeUPay(UPay c) throws Exception {
		encoding_length++;

		compSizeInt(c.getValue());
	}

	public void encodeCertificate(Certificate c) throws Exception {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;

		encodeString((String) c.getName());

		encode(c.getKey());
	}

	public Certificate decodeCertificate(byte[] in) throws Exception {
		Certificate c = new Certificate();
		encoding_length++;
		c.setName(decodeString(in));
		c.setKey(((PublicKey) decode(in)));
		return c;
	}

	private void compSizeCertificate(Certificate c) throws Exception {
		encoding_length++;

		compSizeString(c.getName());
		compSize(c.getKey());
	}

	public void encodeSignedData(SignedData c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;

		encodeByteArray((byte[]) c.getSigned());
	}

	public SignedData decodeSignedData(byte[] in) throws Exception {
		SignedData c = new SignedData();
		encoding_length++;
		c.setSigned(decodeByteArray(in));
		return c;
	}

	private void compSizeSignedData(SignedData c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getSigned());
	}

	public void encodePayData(PayData c) throws Exception {
		encodeDestination[encoding_length++] = Code.PAYDATA;

		encodeInt((int) c.getAmount());

		encode(c.getChallenge());
	}

	public PayData decodePayData(byte[] in) throws Exception {
		PayData c = new PayData();
		encoding_length++;
		c.setAmount(decodeInt(in));
		c.setChallenge(((Nonce) decode(in)));
		return c;
	}

	private void compSizePayData(PayData c) throws Exception {
		encoding_length++;

		compSizeInt(c.getAmount());
		compSize(c.getChallenge());
	}

	private void encodeEncDataAsymm(EncDataAsymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATAASYMM;
		encodeByteArray(c.encrypted);

		encodeInt(c.plainlength);
		encodeInt(c.enclength);

		encodeByte(c.plainDataType);
	}

	public EncDataAsymm decodeEncDataAsymm(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.ENCDATAASYMM)
			stop();
		EncDataAsymm into = new EncDataAsymm();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		into.encrypted = decodeByteArray(in);
		into.plainlength = decodeInt(in);
		into.enclength = decodeInt(in);
		into.plainDataType = decodeByte(in);
		return into;
	}

	private void compSizeEncDataAsymm(EncDataAsymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	private void encodePublicKey(PublicKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;
		encodeByteArray(c.getModulus());

	}

	public PublicKey decodePublicKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PUBLICKEY)
			stop();
		PublicKey res = new PublicKey();
		res.setKey(decodeByteArray(in));
		return res;
	}

	private void compSizePublicKey(PublicKey c) throws Exception {
		encoding_length++;
		compSizeByteArray(c.getModulus());

	}

	public void encodeNonce(Nonce c) throws Exception {
		encodeDestination[encoding_length++] = Code.NONCE;

		encodeByteArray((byte[]) c.getNonce());
	}

	public Nonce decodeNonce(byte[] in) throws Exception {
		Nonce c = new Nonce();
		encoding_length++;
		c.setNonce(decodeByteArray(in));
		return c;
	}

	private void compSizeNonce(Nonce c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getNonce());
	}

	public void encodePrivateKey(PrivateKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.getModulus());
		encodeByteArray(c.getPrivexponent());

	}

	public PrivateKey decodePrivateKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			stop();
		PrivateKey res = new PrivateKey();
		res.setKey(decodeByteArray(in), decodeByteArray(in));
		return res;
	}

	private void compSizePrivateKey(PrivateKey c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getModulus());
		compSizeByteArray(c.getPrivexponent());

	}

	private void stop() throws TerminalException {
		throw new TerminalException("Coding");
	}
}
