package generated.geldkarteAsymm.terminal.merchant;

//import swt.simulation.SmartCardSimulation;
import swt.terminal.HandleResult;
import swt.util.ByteArray;
import swt.util.HexString;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReader;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReaderException;
import javax.xml.ws.BindingProvider;

public class Merchant {

	private int value;

	private PublicKey bankKey;

	private PublicKey cardKey;

	private Nonce challenge;

	private int collected;

	private PayData pd;

	private Userinterface userinterface;
	public void setUserinterface(Userinterface u) {
		this.userinterface = u;
	}

	// saved prev message
	public byte[] prevMsg;
	public U prevUMsg;

	public final Gate dummyGate = new Gate() {
		@Override
		public void callConstraints() {
			/* do nothing */
		}

		@Override
		public byte[] transmit(Merchant terminal, byte[] apdu, int port)
				throws TerminalException, SWTReaderException {
			return terminal.transmit(apdu, port);
		}

		@Override
		public void finishProtocol(Merchant terminal) {
			/* do nothing */
		}

	};
	public Gate currentGate = dummyGate;

	/**
	 * Resets the current gate to a default dummy gate
	 */
	public void resetGate() {
		currentGate = dummyGate;
	}

	/**
	 * Sets the current gate
	 */
	public void setGate(Gate gate) {
		currentGate = gate;
	}

	public Merchant() throws java.lang.Exception {

		value = 0;
		bankKey = new PublicKey();
		cardKey = new PublicKey();
		challenge = new Nonce();
		collected = 0;
		pd = new PayData();

	}

	private void sendMsg(Message msg) {

		currentGate.callConstraints();
		userinterface.send((U) msg);

	}

	private void sendMsg(Message msg, int port) throws java.lang.Exception {

		sendMsg(msg, port, false, false);

	}
	private void sendMsg(Message msg, int port, Boolean openSessionBeforeSend,
			Boolean closeSessionAfterReceive) throws java.lang.Exception {

		currentGate.callConstraints();

		Message response;

		switch (port) {

			case Ports.Merchant2Geldkarte :
				response = sendCardMsg(msg, port);
				if (response != null)
					processMessage(response);
				break;

			default :
				stop();
		}
	}

	private Message sendCardMsg(Message msg, int port)
			throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode((Codeable) msg, encMsg);
		encMsg = ByteArray.subarray(encMsg, 0, encMsgLen);
		byte[] apdu = ByteArray.append(new byte[]{0, 2, 0, 0, 0}, encMsg);
		byte[] response = currentGate.transmit(this, apdu, port);

		if (response.length == 2) {
			if (!(response[0] == (byte) 0x90 && response[1] == (byte) 0x00))
				throw new TerminalException("Cardexception: "
						+ HandleResult.interpretSW(response));
			return null;
		} else {
			prevMsg = response;

			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(response);
			return (Message) cResponse;
		}
	}

	/**
	 * Processes a raw data stream by trying to decode data to Message.
	 * Needed for testcases.
	 * @param data
	 */
	public void processRawMessage(byte[] data) {
		prevMsg = data;
		try {
			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(data);
			processMessage((Message) cResponse);
		} catch (java.lang.Exception e) {
			System.err.println("Failed decoding data in processRawMessage");
		}

	}

	private void processMessage(Message inmsg) throws java.lang.Exception {
		currentGate.callConstraints();

		switch (inmsg.getCode()) {
			case Code.UPAY :
				processUPay((UPay) inmsg);
				break;
			case Code.SENDKEY :
				processSendKey((SendKey) inmsg);
				break;
			case Code.PAYSIG :
				processPaySig((PaySig) inmsg);
				break;
			default :
				stop();
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	public void process(Message msg) throws TerminalException {
		currentGate.callConstraints();

		try {
			if (msg instanceof U)
				prevUMsg = (U) msg;
			processMessage(msg);
		} catch (java.lang.Exception e) {
			// e.printStackTrace();
			if (e instanceof TerminalException)
				throw (TerminalException) e;
			else
				throw new TerminalException(
						"Error when calling process method: " + e);
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	private final static int MAX_SPLIT_LENGTH = 50;
	private final static int MAX_APDU_LENGTH = 254;
	private final static int APDU_HEADER_LENGTH = 5;
	private final static int APDU_DATA_LENTH = MAX_APDU_LENGTH
			- APDU_HEADER_LENGTH;
	private SWTReader[] reader = new SWTReader[1];

	public void initReader(SWTReader reader, int port) {
		this.reader[port] = reader;
	}

	public byte[] encode(Codeable c) throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode(c, encMsg);
		return ByteArray.subarray(encMsg, 0, encMsgLen);
	}

	public byte[] transmit(byte[] apdu, int port) throws TerminalException,
			SWTReaderException {
		byte[] received = null;
		if (apdu.length <= MAX_APDU_LENGTH) {
			apdu[4] = (byte) (apdu.length - 5);
			received = reader[port].transmit(apdu);
			if (received.length == 2)
				return received;
			received = ByteArray.subarray(received, 0, received.length - 2);
		} else {
			byte[] splittedApdu;
			byte[] apduHeader = ByteArray.subarray(apdu, 0, APDU_HEADER_LENGTH);
			apdu = ByteArray.subarray(apdu, APDU_HEADER_LENGTH);
			int splitCount = apdu.length / APDU_DATA_LENTH;
			if (apdu.length % APDU_DATA_LENTH != 0)
				splitCount++;
			for (int i = 0; i < splitCount; i++) {
				if (i == splitCount - 1) {
					splittedApdu = ByteArray
							.subarray(apdu, i * APDU_DATA_LENTH);
					apduHeader[2] = (byte) 0;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (received.length == 2)
						return received;
					received = ByteArray.subarray(received, 0,
							received.length - 2);
				} else {
					splittedApdu = ByteArray.subarray(apdu,
							i * APDU_DATA_LENTH, APDU_DATA_LENTH);
					apduHeader[2] = (byte) 1;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (!(received[0] == (byte) 0x90 && received[1] == (byte) 0x00)) {
						throw new TerminalException(
								"Bad answer from card: part of a long APDU was not answered with OK (0x9000).");
					}
				}
			}
		}

		for (int i = 0; i < MAX_SPLIT_LENGTH; i++) {
			if (received[0] == 0) {
				received = ByteArray.subarray(received, 1);
				return received;
			} else if (received[0] == 1) {
				received = ByteArray.subarray(received, 1);
				byte[] getMore = new byte[]{0, 6, 0, 0, 0, 1, 1};
				byte[] temp_received = reader[port].transmit(getMore);

				temp_received = ByteArray.subarray(temp_received, 0,
						temp_received.length - 2);
				received = ByteArray.append(temp_received[0], received);
				temp_received = ByteArray.subarray(temp_received, 1);
				received = ByteArray.append(received, temp_received);
			} else
				throw new TerminalException(
						"Bad answer from card: answer to a long APDU must begin with 0 or 1.");
		}
		return received;
	}

	/**
	 * This method initializes the card with the given parameters that correspond 
	 * to the card attributes. A null value will be ignored by the card.
	 * @throws Exception
	 */
	public void initCard_Geldkarte(int port, int balance, SignedData cert,
			PrivateKey privKey, Certificate pubKey, byte state)
			throws java.lang.Exception {
		byte cardCode = Code.GELDKARTE;
		byte[] apdu = {0, 4, 0, 0, 0, cardCode};
		apdu = ByteArray.append(
				apdu,
				ByteArray.append(Code.SHORT,
						ByteArray.toByteArray((short) balance)));
		apdu = ByteArray.append(apdu, (cert != null)
				? encode(cert)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (privKey != null)
				? encode(privKey)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (pubKey != null)
				? encode(pubKey)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, new byte[]{Code.BYTE, state});
		transmit(apdu, port);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUPay(UPay inmsg) throws java.lang.Exception {
		synchronized (this) {
			int val = inmsg.getValue();
			setValue(val);
			sendMsg(new GetKey(), Ports.Merchant2Geldkarte);
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processSendKey(SendKey inmsg) throws java.lang.Exception {
		synchronized (this) {
			Certificate c = inmsg.getCert();
			SignedData s = inmsg.getSig();
			if (SignedData.verify(bankKey, s, c)) {
				cardKey.copy(c.getKey());
				challenge.copy(Nonce.generateNonce());
				pd.copy(new PayData(value, challenge));
				EncDataAsymm e = EncDataAsymm.encrypt(cardKey, pd);
				sendMsg(new Pay(e), Ports.Merchant2Geldkarte);
			} else {
				stop();
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processPaySig(PaySig inmsg) throws java.lang.Exception {
		synchronized (this) {
			SignedData s = inmsg.getSig();
			if (SignedData.verify(cardKey, s, pd)) {
				setCollected((collected + value));
			} else {
				stop();
			}
		}
	}

	public int getValue() {
		return value;
	}

	public PublicKey getBankKey() {
		return bankKey;
	}

	public PublicKey getCardKey() {
		return cardKey;
	}

	public Nonce getChallenge() {
		return challenge;
	}

	public int getCollected() {
		return collected;
	}

	public PayData getPd() {
		return pd;
	}

	public void setValue(int value) {
		this.value = value;
	}

	public void setBankKey(PublicKey bankKey) {
		this.bankKey = bankKey;
	}
	public void setCardKey(PublicKey cardKey) {
		this.cardKey = cardKey;
	}
	public void setChallenge(Nonce challenge) {
		this.challenge = challenge;
	}
	public void setCollected(int collected) {
		this.collected = collected;
	}

	public void setPd(PayData pd) {
		this.pd = pd;
	}

	protected void stop() throws TerminalException {
		currentGate.callConstraints();
		throw new TerminalException("Merchant");
	}

}
